#include "main.h"

extern unsigned pid;
extern unsigned iterations;
extern unsigned initial_nodes;
extern unsigned backoff;
extern unsigned backoff_base;
extern private_t private;
extern shared_mem_t* smp;

void
init_private()
{
  private.node = 1 + initial_nodes + pid;
  private.value = 1 + initial_nodes + (pid * iterations);

}

void
init_memory()
{
}

unsigned
new_node()
{
  return private.node;
}

void
reclaim(unsigned node)
{
  private.node = node;
}

void
init_queue()
{
  unsigned i;

  /* initialize queue */
  smp->head.con = NULL;
  smp->tail.con = NULL;
  
  /* initialize avail list */
  for (i=1; i<MAX_NODES; i++) {
    smp->nodes[i].next.con = i+1;
  }
  smp->nodes[MAX_NODES].next.con = NULL;
  
  /* initialize queue contents */
  if (initial_nodes > 0) {
    for (i=2; i<initial_nodes+2; i++) {
      smp->nodes[i].value = i;
      smp->nodes[i].next.con = i+1;
    }
    smp->nodes[initial_nodes].next.con = NULL;
    smp->head.con = 1;
    smp->tail.con = initial_nodes;
  }
}

void
enqueue(unsigned val)
{
  unsigned node;
  pointer_t tail;

  node = new_node();
  smp->nodes[node].value = val;
  smp->nodes[node].next.con = NULL;

  tail.con = fas(&smp->tail, node);
  if (tail.con == NULL) {
    smp->head.con = node;
  } else {
    smp->nodes[tail.con].next.con = node;
  }
}

unsigned
dequeue()
{
  unsigned value;
  unsigned success;
  pointer_t head;
  pointer_t tail;
  pointer_t next;

  backoff = backoff_base;
  while (1) {
    head.con = fas(&smp->head, NULL);
    if (head.con != NULL) {
      break;
    } else if (smp->tail.con == NULL) {
      assert(0);
      return NULL;
    }
    backoff_delay();
  }
  next.con = smp->nodes[head.con].next.con;
  if (next.con == NULL) {
    if (cas(&smp->tail, head.con, NULL) == FALSE) {
      backoff = backoff_base;
      while (smp->nodes[head.con].next.con == NULL) {
	backoff_delay();
      }
      smp->head.con = smp->nodes[head.con].next.con;
    }
  } else {
    smp->head.con = next.con;
  }
  value = smp->nodes[head.con].value;
  reclaim(head.con);
  return value;
}

