#include "main.h"

struct tms tim;
struct tms tim1;

usptr_t *Handle;
barrier_t *Barrier;

int shmid;

unsigned pid;
unsigned procs = 1;
unsigned iterations = 1;
unsigned debug = 0;
unsigned initial_nodes = 0;
char* Fd_name = "out";
shared_mem_t *smp;

void 
time_test ()
{
  unsigned i;
  struct tms time_val;
  clock_t t1, t2;
  unsigned short node;
  unsigned val;
  log_t* plog;
  unsigned ind;

  srandom (pid*pid);
  if(pid==0) {
    smp->lock = 0;

    /* initialize queue */
    smp->head.sep.ptr = NULL;
    smp->head.sep.count = 0;
    smp->tail.sep.ptr = NULL;
    smp->tail.sep.count = 0;

    /* initialize avail list */
    for (i=1; i<MAX_NODES; i++) {
      smp->nodes[i].next.sep.ptr = i+1;
      smp->nodes[i].next.sep.count = 0;
      smp->nodes[i].status.sep.count = 0;
    }
    smp->nodes[MAX_NODES].next.sep.ptr = NULL;
    smp->nodes[MAX_NODES].next.sep.count = 0;

    /* initialize queue */
    if (initial_nodes > 0) {
      for (i=1; i<initial_nodes+1; i++) {
	smp->nodes[i].data = i;
	smp->nodes[i].next.sep.ptr = i+1;
      } 
      smp->nodes[initial_nodes].next.sep.ptr = 1;
      smp->head.sep.ptr = 1;
      smp->tail.sep.ptr = initial_nodes;
    }
  }
  node = 1 + initial_nodes + pid;
  barrier(Barrier,procs);
  if (pid==0) {
    t1=times(&time_val);
  }
  for (i=0;i<iterations;i++)
  {
    val = 1 + i + (pid * iterations);
    smp->nodes[node].data = val;
    if(debug) {
      if (smp->lock) exit(1);
      ind = fai(&smp->index);
      if (ind == LOG_SIZE) {
        faa(&smp->index, -LOG_SIZE);
      }
      plog = &smp->log[ind % LOG_SIZE];
      plog->pid = pid;
      plog->op = PRE_ENQ_C;
      plog->value = val;
    }
    enqueue(node);
    if(debug) {
      if (smp->lock) exit(1);
      ind = fai(&smp->index);
      if (ind == LOG_SIZE) {
        faa(&smp->index, -LOG_SIZE);
      }
      plog = &smp->log[ind % LOG_SIZE];
      plog->pid = pid;
      plog->op = POST_ENQ_C;
    }
    if (debug) {
      if (smp->lock) exit(1);
      ind = fai(&smp->index);
      if (ind == LOG_SIZE) {
        faa(&smp->index, -LOG_SIZE);
      }
      plog = &smp->log[ind % LOG_SIZE];
      plog->pid = pid;
      plog->op = PRE_DEQ_C;
    }
    node = dequeue();
    assert(node != NULL);
    if (debug) {
      val = smp->nodes[node].data;
      if (smp->lock) exit(1);
      ind = fai(&smp->index);
      if (ind == LOG_SIZE) {
        faa(&smp->index, -LOG_SIZE);
      }
      plog = &smp->log[ind % LOG_SIZE];
      plog->pid = pid;
      plog->op = POST_DEQ_C;
      plog->value = val;
    }
  }
  barrier(Barrier,procs);
  if(pid==0)
  {
    t2=times(&time_val);
    printf("procs=%d  iterations=%d  time=%.3f seconds",
	   procs, iterations,(t2-t1)/(double)HZ);
    printf("\n");
    fflush(stdout);
  }
}

void main_task()
{
  time_test();
}

void  setup_shmem()
{
    shmid = shmget(IPC_PRIVATE, sizeof(shared_mem_t), 511);
    assert(shmid != -1);
    smp = (shared_mem_t *)shmat(shmid, 0, 0);
    assert((int)smp != -1);
}

void my_m_fork(void (*func)(),int num_procs)
{
int i;
    
  for (i=1;i<num_procs;i++) {
    pid = i;
    if(fork()==0) /* Child */ {
      (*func)(); /* Call the program */
      return;
    }
  }
  pid=0;
  (*func)(); /* Call the program */
}

main(int argc,char **argv)
{
  parse_args(argc, argv);
  setup_shmem();
  Handle = usinit("/tmp/foo_barrier");
  Barrier = new_barrier(Handle);
  init_barrier(Barrier);
  my_m_fork(main_task, procs); 
  shmctl(shmid, IPC_RMID, smp);
  exit(0);
}

