#include "main.h"

extern unsigned pid;
extern unsigned debug;
extern shared_mem_t* smp;

void
enqueue(unsigned short node)
{
  unsigned success;
  pointer_t anchor;
  pointer_t next;
  pointer_t status;
  log_t* plog;
  unsigned ind;

  for (success = FALSE; success == FALSE; ) {
    /* Test for empty queue */
    anchor.con = smp->anchor.con;
    if (anchor.sep.ptr == NULL) {
      /* Empty queue */
      smp->nodes[node].next.sep.ptr = node;      
      smp->nodes[node].status.sep.ptr = UNLOCKED_S;
      if (debug) {
	if (smp->lock) exit(1);
	ind = fai(&smp->index);
	if (ind == LOG_SIZE) {
	  faa(&smp->index, -LOG_SIZE);
	}
	plog = &smp->log[ind % LOG_SIZE];
	plog->pid = pid;
	plog->op = ENQ_EMPTY_C;
	plog->old.con = anchor.con;
	plog->new.con = MAKE_LONG(node, anchor.sep.count+1);
      }
      success = cas(&smp->anchor,
		    anchor.con, 
		    MAKE_LONG(node, anchor.sep.count+1));
    } else {
      /* Test for Enqueue stage 2 */
      status.con = smp->nodes[anchor.sep.ptr].status.con;
      if (status.sep.ptr == LOCKED_S) {
	/* Enqueue statge 2 */
	if (debug) {
	  if (smp->lock) exit(1);
	  ind = fai(&smp->index);
	  if (ind == LOG_SIZE) {
	    faa(&smp->index, -LOG_SIZE);
	  }
	  plog = &smp->log[ind % LOG_SIZE];
	  plog->pid = pid;
	  plog->op = ENQ_ENQ2_C;
	  plog->old.con = status.con;
	  plog->new.con = MAKE_LONG(UNLOCKED_S, status.sep.count+1);
	}
	cas(&smp->nodes[anchor.sep.ptr].status,
	    status.con,
	    MAKE_LONG(UNLOCKED_S, status.sep.count+1));
      } else {
	/* Test for Enqueue statge 1 */
	next.con = smp->nodes[anchor.sep.ptr].next.con;
	status.con = smp->nodes[next.sep.ptr].status.con;
	if (status.sep.ptr == LOCKED_S) {
	  /* Enqueue statge 1 */
	  if (debug) {
	    if (smp->lock) exit(1);
	    ind = fai(&smp->index);
	    if (ind == LOG_SIZE) {
	      faa(&smp->index, -LOG_SIZE);
	    }
	    plog = &smp->log[ind % LOG_SIZE];
	    plog->pid = pid;
	    plog->op = ENQ_ENQ1_C;
	    plog->old.con = anchor.con;
	    plog->new.con = MAKE_LONG(next.sep.ptr, anchor.sep.count+1);
	  }
	  cas(&smp->anchor,
	      anchor.con,
	      MAKE_LONG(next.sep.ptr, anchor.sep.count+1));
	} else {
	  /* Non-empty non-transient queue */
	  smp->nodes[node].next.sep.ptr = next.sep.ptr;
	  smp->nodes[node].status.sep.ptr = LOCKED_S;
	  status.con = smp->nodes[node].status.con;
	  if (debug) {
	    if (smp->lock) exit(1);
	    ind = fai(&smp->index);
	    if (ind == LOG_SIZE) {
	      faa(&smp->index, -LOG_SIZE);
	    }
	    plog = &smp->log[ind % LOG_SIZE];
	    plog->pid = pid;
	    plog->op = ENQ_NON_EMPTY_C;
	    plog->old.con = next.con;
	    plog->new.con = MAKE_LONG(node, next.sep.count+1);
	  }
	  success = cas(&smp->nodes[anchor.sep.ptr].next,
			next.con,
			MAKE_LONG(node, next.sep.count+1));
	  if (success == TRUE) {
	    if (debug) {
	      if (smp->lock) exit(1);
	      ind = fai(&smp->index);
	      if (ind == LOG_SIZE) {
		faa(&smp->index, -LOG_SIZE);
	      }
	      plog = &smp->log[ind % LOG_SIZE];
	      plog->pid = pid;
	      plog->op = ENQ_NON_EMPTY1_C;
	      plog->old.con = anchor.con;
	      plog->new.con = MAKE_LONG(node, anchor.sep.count+1);
	    }
	    if (cas(&smp->anchor,
		    anchor.con,
		    MAKE_LONG(node, anchor.sep.count+1)) == TRUE) {
	      if (debug) {
		if (smp->lock) exit(1);
		ind = fai(&smp->index);
		if (ind == LOG_SIZE) {
		  faa(&smp->index, -LOG_SIZE);
		}
		plog = &smp->log[ind % LOG_SIZE];
		plog->pid = pid;
		plog->op = ENQ_NON_EMPTY2_C;
		plog->old.con = status.con;
		plog->new.con = MAKE_LONG(UNLOCKED_S, status.sep.count+1);
	      }
	      cas(&smp->nodes[node].status,
		  status.con,
		  MAKE_LONG(UNLOCKED_S, status.sep.count+1));
	    }
	  }
	}
      }
    }
  }
}

unsigned short
dequeue()
{
  unsigned success;
  unsigned short node;
  pointer_t anchor;
  pointer_t next;
  pointer_t status;
  log_t* plog;
  unsigned ind;

  for (success = FALSE; success == FALSE; ) {
    /* Test for empty queue */
    anchor.con = smp->anchor.con;
    if (anchor.sep.ptr == NULL) {
      smp->lock = 1;
      fprintf(stderr,"p%d anchor == NULL\n",pid);
      /* Empty queue */
      dump_log();
      assert(anchor.sep.ptr != NULL);
    }
    /* Test for Enqueue stage 2 */
    status.con = smp->nodes[anchor.sep.ptr].status.con;
    if (status.sep.ptr == LOCKED_S) {
      /* Enqueue statge 2 */
      if (debug) {
	if (smp->lock) exit(1);
	ind = fai(&smp->index);
	if (ind == LOG_SIZE) {
	  faa(&smp->index, -LOG_SIZE);
	}
	plog = &smp->log[ind % LOG_SIZE];
	plog->pid = pid;
	plog->op = DEQ_ENQ2_C;
	plog->old.con = status.con;
	plog->new.con = MAKE_LONG(UNLOCKED_S, status.sep.count+1);
      }
      cas(&smp->nodes[anchor.sep.ptr].status,
	  status.con,
	  MAKE_LONG(UNLOCKED_S, status.sep.count+1));
    } else {
      /* Test for one item queue */
      next.con = smp->nodes[anchor.sep.ptr].next.con;
      if (next.sep.ptr == anchor.sep.ptr) {
	/* One-item queue */
	if (debug) {
	  if (smp->lock) exit(1);
	  ind = fai(&smp->index);
	  if (ind == LOG_SIZE) {
	    faa(&smp->index, -LOG_SIZE);
	  }
	  plog = &smp->log[ind % LOG_SIZE];
	  plog->pid = pid;
	  plog->op = DEQ_ONE_ITEM_C;
	  plog->old.con = anchor.con;
	  plog->new.con = MAKE_LONG(NULL, anchor.sep.count+1);
	}
	success = cas(&smp->anchor,
		     anchor.con,
		     MAKE_LONG(NULL, anchor.sep.count+1));
      } else {
	/* Test for Enqueue stage 1 */
	status.con = smp->nodes[next.sep.ptr].status.con;
	if (status.sep.ptr == LOCKED_S) {
	  /* Enqueue statge 1 */
	  if (debug) {
	    if (smp->lock) exit(1);
	    ind = fai(&smp->index);
	    if (ind == LOG_SIZE) {
	      faa(&smp->index, -LOG_SIZE);
	    }
	    plog = &smp->log[ind % LOG_SIZE];
	    plog->pid = pid;
	    plog->op = DEQ_ENQ1_C;
	    plog->old.con = anchor.con;
	    plog->new.con = MAKE_LONG(next.sep.ptr, anchor.sep.count+1);
	  }
	  cas(&smp->anchor,
	      anchor.con,
	      MAKE_LONG(next.sep.ptr, anchor.sep.count+1));
	} else {
	  /* Multi-item non-transient queue */
	  if (debug) {
	    if (smp->lock) exit(1);
	    ind = fai(&smp->index);
	    if (ind == LOG_SIZE) {
	      faa(&smp->index, -LOG_SIZE);
	    }
	    plog = &smp->log[ind % LOG_SIZE];
	    plog->pid = pid;
	    plog->op = DEQ_MULTI_ITEM_C;
	    plog->old.con = next.con;
	    plog->new.con = MAKE_LONG(smp->nodes[next.sep.ptr].next.sep.ptr,
				      next.sep.count+1);
	  }
	  success = cas(&smp->nodes[anchor.sep.ptr].next,
			next.con,
			MAKE_LONG(smp->nodes[next.sep.ptr].next.sep.ptr,
				  next.sep.count+1));
	}
      }
    }
  }
  return next.sep.ptr;
}

