#include "main.h"

extern unsigned pid;
extern unsigned iterations;
extern unsigned initial_nodes;
extern unsigned backoff;
extern unsigned backoff_base;
extern private_t private;
extern shared_mem_t* smp;

node_t* safe_read(node_t** ptr);
void release(node_t* pnode);
void enqueue(unsigned value);

void
init_private()
{
  private.value = 1 + initial_nodes + (pid * iterations);
}

void
init_memory()
{
}

void
init_queue()
{
  unsigned i;

  /* initialize queue */
  smp->head = &smp->nodes[1];
  smp->tail = &smp->nodes[1];
  smp->nodes[1].next = NULL;
  smp->nodes[1].refct = 4;  /* the lowest bit is the claim bit */
  
  /* initialize avail list */
  for (i=2; i<MAX_NODES+1; i++) {
    smp->nodes[i].next = &smp->nodes[i+1];
    smp->nodes[i].refct = 1; /* the claim bit is set, not necessary */
  }
  smp->nodes[MAX_NODES].next = NULL;
  smp->Avail = &smp->nodes[2];
  
  /* initialize queue contents */
  for (i=2; i<initial_nodes+2; i++) {
    enqueue(i);
  }
  smp->allocated = 0;
  smp->reclaimed = 0;
}

unsigned
dec_tas(unsigned* ptr)
{
  unsigned old;
  unsigned new;

  do {
    old = *ptr;
    new = old - 2;
    if (new == 0) {
      new = 1;
    }
  } while (cas(ptr, old, new) == FALSE);
  return (old == 2);
}

void
clear_bit(unsigned* ptr)
{
  unsigned old;
  unsigned new;

  do {
    old = *ptr;
    new = old - 1;
  } while (cas(ptr, old, new) == FALSE);
}

node_t*
alloc()
{
  node_t* pnode;
  unsigned success;

  for (success = FALSE; success == FALSE; ) {
    pnode = safe_read(&smp->Avail);
    assert(pnode != NULL);
    success = cas(&smp->Avail, pnode, pnode->next);
    if (success == FALSE) {
      release(pnode);
    }
  }
  faa(&smp->allocated, 1);
  return pnode;
}

void
reclaim(node_t* pnode)
{
  node_t* old;

  do {
    old = smp->Avail;
    pnode->next = old;
  } while (cas(&smp->Avail, old, pnode) == FALSE);
  faa(&smp->reclaimed, 1);
}

node_t*
new_node()
{
  node_t* pnode;

  pnode = alloc();
  faa(&pnode->refct, 2);  /* adjust */
  clear_bit(&pnode->refct);
  return pnode;
}

void
release(node_t* pnode)
{
  node_t* ptr;
  node_t* next;

  ptr = pnode;
  while(1) {
    if (dec_tas(&ptr->refct) == FALSE) {
      return;
    }
    next = ptr->next;
    reclaim(ptr);
    ptr = next;
  }
}

node_t*
safe_read(node_t** ptr)
{
  node_t* pnode;

  while(1) {
    pnode = *ptr;
    if (pnode == NULL) {
      return NULL;
    }
    faa(&pnode->refct, 2);
    if (pnode == *ptr) {
      return pnode;
    }
    release(pnode);
  }
}

void
enqueue(unsigned value)
{
  node_t* pnode;
  node_t* ptr;
  node_t* old;
  node_t* r;

  pnode = new_node();
  pnode->value = value;
  pnode->next = NULL;

  ptr = safe_read(&smp->tail);
  old = ptr; /* old = safe_read(&ptr); */
  backoff = backoff_base;
  do {
    while (ptr->next != NULL) {
      r = ptr;
      ptr = safe_read(&ptr->next);
      release(r);
    }
  } while (cas(&ptr->next, NULL, pnode) == FALSE);
  release(ptr);
  if (cas(&smp->tail, old, pnode) == TRUE) {
    /* release(old); */
    release(old);
  } else {
    /* release(old); */
    release(pnode);
  }
}

#ifdef AAA
void
enqueue(unsigned value)
{
  node_t* pnode;
  node_t* ptr;
  unsigned success;

  pnode = new_node();
  pnode->value = value;
  pnode->next = NULL;

  backoff = backoff_base;
  for (success = FALSE; success == FALSE;) {
    ptr = safe_read(&smp->tail);
    success = cas(&ptr->next, NULL, pnode);
    if (success == FALSE) {
      faa(&ptr->next->refct);
      if (cas(&smp->tail, ptr, ptr->next) == TRUE) {
	release(ptr);
      } else {
	release(ptr->next);
      }
      backoff_delay();
    }
  }
  release(ptr);
  if (cas(&smp->tail, ptr, pnode) == TRUE) {
    release(ptr);
  } else {
    release(pnode);
  }
}
#endif

unsigned
dequeue()
{
  unsigned success;
  unsigned value;
  node_t* pnode;
  node_t* next;

  backoff = backoff_base;
  for (success = FALSE; success == FALSE; ) {
    pnode = safe_read(&smp->head);
    next = pnode->next;
    if (next == NULL) {
      release(pnode);
      return NULL;
    }
    faa(&next->refct, 2);
    success = cas(&smp->head, pnode, next);
    if (success == FALSE) {
      release(next);
      backoff_delay();
    }
    release(pnode);
  }
  value = pnode->next->value;
  release(pnode);
  return value;
}

