#include "main.h"

extern unsigned pid;
extern unsigned iterations;
extern unsigned initial_nodes;
extern private_t private;
extern shared_mem_t* smp;


void
init_private()
{
  private.node = 1 + initial_nodes + pid;
  private.value = 1 + initial_nodes + (pid * iterations);
}

void
init_memory()
{
}

unsigned
new_node()
{
  return private.node;
}

void
reclaim(unsigned node)
{
  private.node = node;
}

void
init_queue()
{
  unsigned i;

  /* init lock */
  smp->lock = 0;

  /* initialize empty queue */
  smp->head = NULL;
  smp->tail = NULL;
  
  /* initialize queue contents */
  if (initial_nodes > 0) {
    smp->head = 1;
    smp->tail = initial_nodes;
    for (i=1; i<initial_nodes+1; i++) {
      smp->nodes[i].value = i;
      smp->nodes[i].next = i+1;
    }
    smp->nodes[initial_nodes].next = NULL;
  }
}

void
enqueue(unsigned val)
{
  unsigned node;

  node = new_node();
  smp->nodes[node].value = val;
  smp->nodes[node].next = NULL;
  tts_acq(&smp->lock); /* native_acquire(); */
  if (smp->tail == NULL) {
    smp->head = node;
  } else {
    smp->nodes[smp->tail].next = node;
  }
  smp->tail = node;
  smp->lock = 0; /* native_release(); */
}

unsigned
dequeue()
{
  unsigned val;
  unsigned node;

  tts_acq(&smp->lock); /* native_acquire(); */
  if (smp->head == NULL) {
    node = NULL;
  }
  node = smp->head;
  smp->head = smp->nodes[node].next;
  if (smp->head == NULL) {
    smp->tail = NULL;
  }
  smp->lock = 0; /* native_release(); */
  assert(node != NULL);
  val = smp->nodes[node].value;
  reclaim(node);
  return val;
}
